/*
    Luftdruckmessung mit Arduino UNO und Adafruit Datalogging Shield

    Wasja Kleider
    Juli 2015
    Universität Heidelberg
    
    Speicherpfad auf SD-Karte: "P_LOG_XX.DAT" (XX=fortlaufende Nummer 00-99)

    Datenformat: "t (ms) \t t_RTC \t T (C) \t p (mbar)"
    Beispiel:    "73244  2015/7/17 11:36:26  26,60 1002,87" (Tab-separierte Spalten)
    
    Grundlage dieses Sketches ist ein Beispielprogramm von Mike Grusin (Version 10 vom 24.10.2013):
    SFE_BMP180 (https://www.sparkfun.com/products/11824) 
*/

// Einstellungen für Messungen NUR hier ändern:
#define measureInterval 2000  // Zeitabstand der Einzelmessungen in Millisekunden, Standard=2000 (2s).
#define baudRate 115200       // Baudrate für die Serielle Kommunikation, Standard = 115200.

// Für das Standardprogramm: Ab hier keine Änderungen mehr vornehmen!

// nötige Libraries laden
#include "Wire.h"   // I2C-Kommunikation (RealTimeClock)
#include "RTClib.h" // RealTimeClock
#include "SFE_BMP180.h"// Drucksensor (BMP180 von Bosch/Adafruit Version)
#include "SPI.h"    // SPI-Kommunikation für SD-Karte
#include "SD.h"     // SD-Karte

// Variablen für SD-Kommunikation
const int CS = 10; // SPI-Chip Select
// weitere SPI-Pins: MOSI 11, MISO 12, CLK 13 -- auf Datalogging Shield vorgegeben

// Datei, in der die Daten gespeichert werden
File logFile;

// Pinbelegungen für das Bedienungsfeld
const int errorLED = 7;     // LED fuer Anzeige eines Fehlers bei Initialisierung oder Messung
const int measureLED = 6;   // LED fuer Anzeige einer Messwertaufnahme
const int startButton = 4;  // Button für Messstart
const int stopButton = 5;   // Button für Messstopp

// Variablen für die Messung
unsigned long currentTimestamp = 0; // Zeitmarke fuer Messungen
bool measurementInProgress = false;
char filenamesave[] = "TESTTEST.DAT";

// Echtzeituhr
RTC_DS1307 RTC;

// Drucksensor
SFE_BMP180 pressureSensor;


void setup()
{
  led_both();
  Serial.begin(baudRate);
  Serial.println("Programmstart.");
  
  // Einrichten der LED und der Buttons
  pinMode(errorLED, OUTPUT);
  pinMode(measureLED, OUTPUT); 
  pinMode(startButton, INPUT_PULLUP); // Pullup-Widerstand aktiv, sodass LOW einem gedrueckten Knopf entspricht
  pinMode(stopButton, INPUT_PULLUP);  // s.o.
  
  // Drucksensor initialisieren.
  if (pressureSensor.begin()) {
    Serial.println("BMP180 initialisiert.");
  }
  else
  {
    error("Fehler Initialisierung BMP180.");
  }
  
  // Kommunikation mit der SD-Karte aufbauen  
  pinMode(CS, OUTPUT); // Chip Select als Output
  if (!SD.begin(CS)) {
    error("Keine SD-Karte erkannt."); // Fehlermeldung anzeigen und Programm anhalten
  }
  Serial.println("SD-Karte initialisiert.");
  
  // Erstelle ein neues Logfile und öffne es.
  char filename[] = "P_LOG_00.DAT";
  for (uint8_t i = 0; i < 100; i++) {
    filename[6] = i/10 + '0';
    filename[7] = i%10 + '0';
    if (! SD.exists(filename)) {
      // Nur öffnen, falls die Datei noch nicht existiert.
      logFile = SD.open(filename, FILE_WRITE);
      
      // Speichere auch den endgültigen Dateiname, damit Daten gespeichert werden können.
      for (int j = 0; j<14; j++) {
        filenamesave[j] = filename[j];
      }
      break;  // Dann Schleife beenden.
    }
  }
  if(!logFile) {
    error("Log-Datei konnte nicht angelegt werden. (bis 99?)"); // Evtl. alte Logfiles löschen!
  }
  else {
    Serial.print("Log-Datei: ");
    Serial.println(filename);  
  }
  
  
  // Starte die Kommunikation mit der Echtzeituhr
  Wire.begin();  
  if (!RTC.begin()) {
    logFile.println("RTC-Fehler.");
    error("RTC-Fehler.");
  }
  
  // Schreibe eine Kopfzeile in die Datei
  logFile.println("t (ms) \t t_RTC \t T (C) \t p (mbar)");
  
  // Schließe das Logfile, damit die bisherigen Daten gespeichert werden.
  logFile.close();
  
  // Signalisiere Ende der Initialisierung durch Ausschalten beider LED
  delay(1000);
  led_none();
}

void loop()
{ 
  // Falls keine Messung läuft, warte bis der Start-Button gedrückt wird.
  if(!measurementInProgress) {
    while(digitalRead(startButton)) {
        // Start-Button wurde nicht gedrückt, also warte.
    }
    measurementInProgress = true;
    Serial.println("Messung gestartet.");
  }
  // Falls eine Messung läuft und der Stop-Button gedrückt wird, halte die Messung an.
  else if(measurementInProgress) {
    if (!digitalRead(stopButton)) {
      // Button ist auf LOW, also gedrueckt
      measurementInProgress = false;
      Serial.println("Messung beendet.");
      led_both();
      delay(500);
      led_none();
      delay(500);
      led_both();
      delay(500);
      led_none();
      return; // Beende die Schleife vorzeitig, sodass auf ein erneutes Startsignal gewartet wird.
    }
  }
  
  // Nimm die gewünschten Zeiten und Messdaten auf.
  led_measure();
  currentTimestamp = millis();
  DateTime now = RTC.now();
  char timeToWait;
  double T, P;
  String stringT = "";
  String stringP = "";
  
  // Starte eine Temperaturmessung. (IMMER vor Druckmessung!)
  timeToWait = pressureSensor.startTemperature();
  if (timeToWait != 0) // T-Messstart erfolgreich.
  {
    delay(timeToWait); // Warten, bis Messstart vollständig.
    
    // Führe die Temperaturmessung durch.
    timeToWait = pressureSensor.getTemperature(T); // Lies das T-Messergebnis aus.
    if (timeToWait != 0) // T-Messung erfolgreich gestartet.
    {
      delay(timeToWait); // Warten, bis T-Messergebnis vollständig.
      
      Serial.print("T=");
      Serial.print(T,2);
      Serial.println("'C");
      
      stringT += T;
      stringT.replace(".", ","); // Ersetze den Dezimalpunkt durch ein Komma.
      
      // Starte eine Druckmessung.
      timeToWait = pressureSensor.startPressure(3);
      if (timeToWait != 0) // P-Messstart erfolgreich. 
      {
        delay(timeToWait); // Warten, bis Messstart vollständig.
        timeToWait = pressureSensor.getPressure(P,T); // Lies das P-Messergebnis aus.
        if (timeToWait != 0) // P-Messung erfolgreich gestartet.
        {
          Serial.print("P=");
          Serial.print(P,2);
          Serial.println("mbar.");
          
          stringP += P;
          stringP.replace(".", ","); // Ersetze den Dezimalpunkt durch ein Komma.
        }
        else {
          error("Fehler bei P-Messung.");
        }
      }
      else 
      {
        error("Fehler bei P-Messstart.");
      }
    }
    else 
    {
      error("Fehler bei T-Messung.");
    }
  }
  else
  {
    error("Fehler bei T-Messstart.");
  }
  
   
  // Öffne das Logfile.
  logFile = SD.open(filenamesave, FILE_WRITE);
  if(!logFile) {
    error("Logfile-Speicherung nicht möglich.");
  }
  
  // Schreibe alle Daten in das Logfile.
  logFile.print(currentTimestamp);
  logFile.print("\t");
  logFile.print(makeTimestamp(now));
  logFile.print("\t");
  logFile.print(stringT);
  logFile.print("\t");
  logFile.println(stringP);
  logFile.close();

  // Messwertaufnahme beenden
  delay(5); // Warte kurz, bevor die Status-LEDs ausgeschaltet werden.
  led_none();

  // Warte, bis der nächste Messwert aufgenommen werden soll.
  while(millis() < (currentTimestamp+measureInterval)) 
  {
    // Falls während der Wartezeit der Stop-Button gedrückt wird, nicht länger warten!
    if(!digitalRead(stopButton)) return;
  }
}


// Schalte beide LED ein.
void led_both() {
  digitalWrite(errorLED, HIGH); // Fehler-LED einschalten
  digitalWrite(measureLED, HIGH); // Messungs-LED einschalten
}

// Schalte nur die Error-LED ein.
void led_error() {
  digitalWrite(errorLED, HIGH); // Fehler-LED einschalten
  digitalWrite(measureLED, LOW); // Messungs-LED ausschalten
}

// Schalte nur die Mess-LED ein.
void led_measure() {
  digitalWrite(errorLED, LOW); // Fehler-LED ausschalten
  digitalWrite(measureLED, HIGH); // Messungs-LED einschalten  
}

// Schalte beide LED aus.
void led_none() {
  digitalWrite(errorLED, LOW); // Fehler-LED ausschalten
  digitalWrite(measureLED, LOW); // Messungs-LED ausschalten
}


// Überführe die aktuelle Zeit in einen lesbaren String des Formats "YYYY/MM/DD HH:MM:SS"
String makeTimestamp(DateTime time) {
  String returnString = "";
  
  returnString += (String) time.year();
  returnString += "/";
  returnString += (String) time.month();
  returnString += "/";
  returnString += (String) time.day();
  returnString += " ";
  returnString += (String) time.hour();
  returnString += ":";
  returnString += (String) time.minute();
  returnString += ":";
  returnString += (String) time.second();
  
  return returnString;
}


// Falls ein Fehler im Programmablauf auftritt, sende eine Fehlernachrit an die
// serielle Schnittstelle, schalte die Fehler-LED ein und halte das Programm an.
void error(char *str)
{
  Serial.print("Fehler: ");
  Serial.println(str);
  led_error();
  while(1);
}
